/*
 * Copyright (C) 2007 Google, Inc.
 * Copyright (c) 2008-2009, Code Aurora Forum. All rights reserved.
 * Author: Brian Swetland <swetland@google.com>
 *
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */






















#include <linux/platform_device.h>
#include <linux/gpio_event.h>

#include <asm/mach-types.h>


#include <linux/gpio.h>
#include <linux/wakelock.h>

#include <linux/delay.h>
#include <linux/hrtimer.h>
#include <mach/hs_io_ctl_a.h>

static struct wake_lock key_wake_lock;

/* don't turn this on without updating the ffa support */
#define SCAN_FUNCTION_KEYS 0

/* FFA:
 36: KEYSENSE_N(0)
 37: KEYSENSE_N(1)
 38: KEYSENSE_N(2)
 39: KEYSENSE_N(3)
 40: KEYSENSE_N(4)

 31: KYPD_17
 32: KYPD_15
 33: KYPD_13
 34: KYPD_11
 35: KYPD_9
 41: KYPD_MEMO
*/












static unsigned int keypad_row_gpios[] = {
	                         GPIO_KEYSENSE_N_0, GPIO_KEYSENSE_N_1,
                             GPIO_KEYSENSE_N_2, GPIO_KEYSENSE_N_3,
                             GPIO_KEYSENSE_N_4, GPIO_KEYSENSE_N_5,
                             GPIO_KEYSENSE_N_6 };
static unsigned int keypad_col_gpios[] = { 
                             GPIO_KYPD_0, GPIO_KYPD_1, GPIO_KYPD_2,
                             GPIO_KYPD_3, GPIO_KYPD_4, GPIO_KYPD_5,
                             GPIO_KYPD_6, GPIO_KYPD_7 };



static unsigned int keypad_row_gpios_8k_ffa[] = {31, 32, 33, 34, 35, 36};
static unsigned int keypad_col_gpios_8k_ffa[] = {38, 39, 40, 41, 42};

#define KEYMAP_INDEX(row, col) ((row)*ARRAY_SIZE(keypad_col_gpios) + (col))

#define KEYMAP_INDEX_MAX (ARRAY_SIZE(keypad_col_gpios)*ARRAY_SIZE(keypad_row_gpios))

#define FFA_8K_KEYMAP_INDEX(row, col) ((row)* \
				ARRAY_SIZE(keypad_col_gpios_8k_ffa) + (col))


#define KEYMAP_ARRAYSIZE (ARRAY_SIZE(keypad_col_gpios) * ARRAY_SIZE(keypad_row_gpios))





static const unsigned short keypad_keymap[KEYMAP_ARRAYSIZE] = {


	[KEYMAP_INDEX(0, 0)] = KEY_Q,
	[KEYMAP_INDEX(0, 1)] = KEY_W,
	[KEYMAP_INDEX(0, 2)] = KEY_E,
	[KEYMAP_INDEX(0, 3)] = KEY_R,
	[KEYMAP_INDEX(0, 4)] = KEY_T,
	[KEYMAP_INDEX(0, 5)] = KEY_Y,
	[KEYMAP_INDEX(0, 6)] = KEY_U,
	[KEYMAP_INDEX(0, 7)] = KEY_WIMAX,

	[KEYMAP_INDEX(1, 0)] = KEY_X,
	[KEYMAP_INDEX(1, 1)] = KEY_C,
	[KEYMAP_INDEX(1, 2)] = KEY_H,
	[KEYMAP_INDEX(1, 3)] = KEY_G,
	[KEYMAP_INDEX(1, 4)] = KEY_F,
	[KEYMAP_INDEX(1, 5)] = KEY_EMAIL,
	[KEYMAP_INDEX(1, 6)] = KEY_WIMAX,
	[KEYMAP_INDEX(1, 7)] = KEY_I,

	[KEYMAP_INDEX(2, 0)] = KEY_DOT,	 
	[KEYMAP_INDEX(2, 1)] = KEY_D,
	[KEYMAP_INDEX(2, 2)] = KEY_L,
	[KEYMAP_INDEX(2, 3)] = KEY_J,
	[KEYMAP_INDEX(2, 4)] = KEY_N,
	[KEYMAP_INDEX(2, 5)] = KEY_S,
	[KEYMAP_INDEX(2, 6)] = KEY_WIMAX,
	[KEYMAP_INDEX(2, 7)] = KEY_O,

	[KEYMAP_INDEX(3, 0)] = KEY_K,
	[KEYMAP_INDEX(3, 1)] = KEY_B,
	[KEYMAP_INDEX(3, 2)] = KEY_Z,
	[KEYMAP_INDEX(3, 3)] = KEY_M,
	[KEYMAP_INDEX(3, 4)] = KEY_SLASH,
	[KEYMAP_INDEX(3, 5)] = KEY_A,
	[KEYMAP_INDEX(3, 6)] = KEY_WIMAX,
	[KEYMAP_INDEX(3, 7)] = KEY_P,

	[KEYMAP_INDEX(4, 0)] = KEY_V,
	[KEYMAP_INDEX(4, 1)] = KEY_VOLUMEUP,
	[KEYMAP_INDEX(4, 2)] = KEY_VOLUMEDOWN,
	[KEYMAP_INDEX(4, 3)] = KEY_ENTER,
	[KEYMAP_INDEX(4, 4)] = KEY_SPACE,
	[KEYMAP_INDEX(4, 5)] = KEY_WIMAX,
	[KEYMAP_INDEX(4, 6)] = KEY_CAPSLOCK,
	[KEYMAP_INDEX(4, 7)] = KEY_WIMAX,

	[KEYMAP_INDEX(5, 0)] = KEY_SEARCH,
	[KEYMAP_INDEX(5, 1)] = KEY_BACK,
	[KEYMAP_INDEX(5, 2)] = KEY_MENU,
	[KEYMAP_INDEX(5, 3)] = KEY_HOME,
	[KEYMAP_INDEX(5, 4)] = 248,	
	[KEYMAP_INDEX(5, 5)] = KEY_SPACE,


	[KEYMAP_INDEX(5, 6)] = KEY_RIGHT,
	[KEYMAP_INDEX(5, 7)] = KEY_UP,

	[KEYMAP_INDEX(6, 0)] = KEY_BACKSPACE,
	[KEYMAP_INDEX(6, 1)] = KEY_WIMAX,
	[KEYMAP_INDEX(6, 2)] = KEY_WIMAX,
	[KEYMAP_INDEX(6, 3)] = 251,	
	[KEYMAP_INDEX(6, 4)] = KEY_WIMAX,
	[KEYMAP_INDEX(6, 5)] = KEY_LEFTALT,	 


	[KEYMAP_INDEX(6, 6)] = KEY_LEFT,
	[KEYMAP_INDEX(6, 7)] = KEY_DOWN

};


static const unsigned short keypad_keymap_surf[ARRAY_SIZE(keypad_col_gpios) *
					  ARRAY_SIZE(keypad_row_gpios)] = {
	[KEYMAP_INDEX(0, 0)] = KEY_5,
	[KEYMAP_INDEX(0, 1)] = KEY_9,
	[KEYMAP_INDEX(0, 2)] = 229,            /* SOFT1 */
	[KEYMAP_INDEX(0, 3)] = KEY_6,
	[KEYMAP_INDEX(0, 4)] = KEY_LEFT,

	[KEYMAP_INDEX(1, 0)] = KEY_0,
	[KEYMAP_INDEX(1, 1)] = KEY_RIGHT,
	[KEYMAP_INDEX(1, 2)] = KEY_1,
	[KEYMAP_INDEX(1, 3)] = 228,           /* KEY_SHARP */
	[KEYMAP_INDEX(1, 4)] = KEY_SEND,

	[KEYMAP_INDEX(2, 0)] = KEY_VOLUMEUP,
	[KEYMAP_INDEX(2, 1)] = KEY_HOME,      /* FA   */
	[KEYMAP_INDEX(2, 2)] = KEY_F8,        /* QCHT */
	[KEYMAP_INDEX(2, 3)] = KEY_F6,        /* R+   */
	[KEYMAP_INDEX(2, 4)] = KEY_F7,        /* R-   */

	[KEYMAP_INDEX(3, 0)] = KEY_UP,
	[KEYMAP_INDEX(3, 1)] = KEY_CLEAR,
	[KEYMAP_INDEX(3, 2)] = KEY_4,
	[KEYMAP_INDEX(3, 3)] = KEY_MUTE,      /* SPKR */
	[KEYMAP_INDEX(3, 4)] = KEY_2,

	[KEYMAP_INDEX(4, 0)] = 230,           /* SOFT2 */
	[KEYMAP_INDEX(4, 1)] = 232,           /* KEY_CENTER */
	[KEYMAP_INDEX(4, 2)] = KEY_DOWN,
	[KEYMAP_INDEX(4, 3)] = KEY_BACK,      /* FB */
	[KEYMAP_INDEX(4, 4)] = KEY_8,

	[KEYMAP_INDEX(5, 0)] = KEY_VOLUMEDOWN,
	[KEYMAP_INDEX(5, 1)] = 227,           /* KEY_STAR */
	[KEYMAP_INDEX(5, 2)] = KEY_MAIL,      /* MESG */
	[KEYMAP_INDEX(5, 3)] = KEY_3,
	[KEYMAP_INDEX(5, 4)] = KEY_7,

#if SCAN_FUNCTION_KEYS
	[KEYMAP_INDEX(6, 0)] = KEY_F5,
	[KEYMAP_INDEX(6, 1)] = KEY_F4,
	[KEYMAP_INDEX(6, 2)] = KEY_F3,
	[KEYMAP_INDEX(6, 3)] = KEY_F2,
	[KEYMAP_INDEX(6, 4)] = KEY_F1
#endif
};

static const unsigned short keypad_keymap_ffa[ARRAY_SIZE(keypad_col_gpios) *
					      ARRAY_SIZE(keypad_row_gpios)] = {
	/*[KEYMAP_INDEX(0, 0)] = ,*/
	/*[KEYMAP_INDEX(0, 1)] = ,*/
	[KEYMAP_INDEX(0, 2)] = KEY_1,
	[KEYMAP_INDEX(0, 3)] = KEY_SEND,
	[KEYMAP_INDEX(0, 4)] = KEY_LEFT,

	[KEYMAP_INDEX(1, 0)] = KEY_3,
	[KEYMAP_INDEX(1, 1)] = KEY_RIGHT,
	[KEYMAP_INDEX(1, 2)] = KEY_VOLUMEUP,
	/*[KEYMAP_INDEX(1, 3)] = ,*/
	[KEYMAP_INDEX(1, 4)] = KEY_6,

	[KEYMAP_INDEX(2, 0)] = KEY_HOME,      /* A */
	[KEYMAP_INDEX(2, 1)] = KEY_BACK,      /* B */
	[KEYMAP_INDEX(2, 2)] = KEY_0,
	[KEYMAP_INDEX(2, 3)] = 228,           /* KEY_SHARP */
	[KEYMAP_INDEX(2, 4)] = KEY_9,

	[KEYMAP_INDEX(3, 0)] = KEY_UP,
	[KEYMAP_INDEX(3, 1)] = 232, /* KEY_CENTER */ /* i */
	[KEYMAP_INDEX(3, 2)] = KEY_4,
	/*[KEYMAP_INDEX(3, 3)] = ,*/
	[KEYMAP_INDEX(3, 4)] = KEY_2,

	[KEYMAP_INDEX(4, 0)] = KEY_VOLUMEDOWN,
	[KEYMAP_INDEX(4, 1)] = KEY_SOUND,
	[KEYMAP_INDEX(4, 2)] = KEY_DOWN,
	[KEYMAP_INDEX(4, 3)] = KEY_8,
	[KEYMAP_INDEX(4, 4)] = KEY_5,

	/*[KEYMAP_INDEX(5, 0)] = ,*/
	[KEYMAP_INDEX(5, 1)] = 227,           /* KEY_STAR */
	[KEYMAP_INDEX(5, 2)] = 230, /*SOFT2*/ /* 2 */
	[KEYMAP_INDEX(5, 3)] = KEY_MENU,      /* 1 */
	[KEYMAP_INDEX(5, 4)] = KEY_7,
};

#define QSD8x50_FFA_KEYMAP_SIZE (ARRAY_SIZE(keypad_col_gpios_8k_ffa) * \
			ARRAY_SIZE(keypad_row_gpios_8k_ffa))

static const unsigned short keypad_keymap_8k_ffa[QSD8x50_FFA_KEYMAP_SIZE] = {

	[FFA_8K_KEYMAP_INDEX(0, 0)] = KEY_VOLUMEDOWN,
	/*[KEYMAP_INDEX(0, 1)] = ,*/
	[FFA_8K_KEYMAP_INDEX(0, 2)] = KEY_DOWN,
	[FFA_8K_KEYMAP_INDEX(0, 3)] = KEY_8,
	[FFA_8K_KEYMAP_INDEX(0, 4)] = KEY_5,

	[FFA_8K_KEYMAP_INDEX(1, 0)] = KEY_UP,
	[FFA_8K_KEYMAP_INDEX(1, 1)] = KEY_CLEAR,
	[FFA_8K_KEYMAP_INDEX(1, 2)] = KEY_4,
	/*[KEYMAP_INDEX(1, 3)] = ,*/
	[FFA_8K_KEYMAP_INDEX(1, 4)] = KEY_2,

	[FFA_8K_KEYMAP_INDEX(2, 0)] = KEY_HOME,      /* A */
	[FFA_8K_KEYMAP_INDEX(2, 1)] = KEY_BACK,      /* B */
	[FFA_8K_KEYMAP_INDEX(2, 2)] = KEY_0,
	[FFA_8K_KEYMAP_INDEX(2, 3)] = 228,           /* KEY_SHARP */
	[FFA_8K_KEYMAP_INDEX(2, 4)] = KEY_9,

	[FFA_8K_KEYMAP_INDEX(3, 0)] = KEY_3,
	[FFA_8K_KEYMAP_INDEX(3, 1)] = KEY_RIGHT,
	[FFA_8K_KEYMAP_INDEX(3, 2)] = KEY_VOLUMEUP,
	/*[KEYMAP_INDEX(3, 3)] = ,*/
	[FFA_8K_KEYMAP_INDEX(3, 4)] = KEY_6,

	[FFA_8K_KEYMAP_INDEX(4, 0)] = 232,		/* OK */
	[FFA_8K_KEYMAP_INDEX(4, 1)] = KEY_SOUND,
	[FFA_8K_KEYMAP_INDEX(4, 2)] = KEY_1,
	[FFA_8K_KEYMAP_INDEX(4, 3)] = KEY_SEND,
	[FFA_8K_KEYMAP_INDEX(4, 4)] = KEY_LEFT,

	/*[KEYMAP_INDEX(5, 0)] = ,*/
	[FFA_8K_KEYMAP_INDEX(5, 1)] = 227,           /* KEY_STAR */
	[FFA_8K_KEYMAP_INDEX(5, 2)] = 230, /*SOFT2*/ /* 2 */
	[FFA_8K_KEYMAP_INDEX(5, 3)] = 229,      /* 1 */
	[FFA_8K_KEYMAP_INDEX(5, 4)] = KEY_7,
};


#define  KEYPRESS_MAX_NUM 0xFFFFFFFF
unsigned int keypad_count[KEYMAP_ARRAYSIZE];



static int keypad_gpio_event_matrix_func(struct gpio_event_input_devs *input_devs,
					  struct gpio_event_info *info,
					  void **data, int func);


/* SURF keypad platform device information */
static struct gpio_event_matrix_info surf_keypad_matrix_info = {


	.info.func	= keypad_gpio_event_matrix_func,



 	.keymap		= keypad_keymap,

	.output_gpios	= keypad_row_gpios,
	.input_gpios	= keypad_col_gpios,
	.noutputs	= ARRAY_SIZE(keypad_row_gpios),
	.ninputs	= ARRAY_SIZE(keypad_col_gpios),
	.settle_time.tv.nsec = 40 * NSEC_PER_USEC,
	.poll_time.tv.nsec = 20 * NSEC_PER_MSEC,
	.flags		= GPIOKPF_LEVEL_TRIGGERED_IRQ | GPIOKPF_DRIVE_INACTIVE |
			  GPIOKPF_PRINT_UNMAPPED_KEYS
};

static struct gpio_event_info *surf_keypad_info[] = {
	&surf_keypad_matrix_info.info
};

static struct gpio_event_platform_data surf_keypad_data = {
	.name		= "surf_keypad",
	.info		= surf_keypad_info,
	.info_count	= ARRAY_SIZE(surf_keypad_info)
};

struct platform_device keypad_device_surf = {
	.name	= GPIO_EVENT_DEV_NAME,
	.id	= -1,
	.dev	= {
		.platform_data	= &surf_keypad_data,
	},
};

/* 8k FFA keypad platform device information */
static struct gpio_event_matrix_info keypad_matrix_info_8k_ffa = {
	.info.func	= gpio_event_matrix_func,
	.keymap		= keypad_keymap_8k_ffa,
	.output_gpios	= keypad_row_gpios_8k_ffa,
	.input_gpios	= keypad_col_gpios_8k_ffa,
	.noutputs	= ARRAY_SIZE(keypad_row_gpios_8k_ffa),
	.ninputs	= ARRAY_SIZE(keypad_col_gpios_8k_ffa),
	.settle_time.tv.nsec = 40 * NSEC_PER_USEC,
	.poll_time.tv.nsec = 20 * NSEC_PER_MSEC,
	.flags		= GPIOKPF_LEVEL_TRIGGERED_IRQ | GPIOKPF_DRIVE_INACTIVE |
			  GPIOKPF_PRINT_UNMAPPED_KEYS
};

static struct gpio_event_info *keypad_info_8k_ffa[] = {
	&keypad_matrix_info_8k_ffa.info
};

static struct gpio_event_platform_data keypad_data_8k_ffa = {
	.name		= "8k_ffa_keypad",
	.info		= keypad_info_8k_ffa,
	.info_count	= ARRAY_SIZE(keypad_info_8k_ffa)
};

struct platform_device keypad_device_8k_ffa = {
	.name	= GPIO_EVENT_DEV_NAME,
	.id	= -1,
	.dev	= {
		.platform_data	= &keypad_data_8k_ffa,
	},
};

/* 7k FFA keypad platform device information */
static struct gpio_event_matrix_info keypad_matrix_info_7k_ffa = {
	.info.func	= gpio_event_matrix_func,
	.keymap		= keypad_keymap_ffa,
	.output_gpios	= keypad_row_gpios,
	.input_gpios	= keypad_col_gpios,
	.noutputs	= ARRAY_SIZE(keypad_row_gpios),
	.ninputs	= ARRAY_SIZE(keypad_col_gpios),
	.settle_time.tv.nsec = 40 * NSEC_PER_USEC,
	.poll_time.tv.nsec = 20 * NSEC_PER_MSEC,
	.flags		= GPIOKPF_LEVEL_TRIGGERED_IRQ | GPIOKPF_DRIVE_INACTIVE |
			  GPIOKPF_PRINT_UNMAPPED_KEYS
};

static struct gpio_event_info *keypad_info_7k_ffa[] = {
	&keypad_matrix_info_7k_ffa.info
};

static struct gpio_event_platform_data keypad_data_7k_ffa = {
	.name		= "7k_ffa_keypad",
	.info		= keypad_info_7k_ffa,
	.info_count	= ARRAY_SIZE(keypad_info_7k_ffa)
};

struct platform_device keypad_device_7k_ffa = {
	.name	= GPIO_EVENT_DEV_NAME,
	.id	= -1,
	.dev	= {
		.platform_data	= &keypad_data_7k_ffa,
	},
};



static struct input_dev *keypad_dev;

#define BUTTONS_POLL_INTERVAL   25  
#define BUTTONS_COUNT_THRESHOLD  2  

static const unsigned short keypad_virtual_keys[] = {
	KEY_END,
	KEY_POWER
};












#define KEY_PUSH_MULTI_MAX  3 

#define KEY_RELEASE 0 
#define KEY_PUSH    1 

#define KEYMAP_ARRAYSIZE_COLUMN ARRAY_SIZE(keypad_col_gpios)
#define KEYMAP_ARRAYSIZE_ROW    ARRAY_SIZE(keypad_row_gpios)

static int key_push_flag[KEYMAP_ARRAYSIZE_COLUMN][KEYMAP_ARRAYSIZE_ROW]; 
static int count[KEYMAP_ARRAYSIZE_COLUMN][KEYMAP_ARRAYSIZE_ROW];



static struct hrtimer key_poll_timer;



static int spacekey_num = 0;





#ifdef KEYPAD_PRESSKEY_DBG
#define KEY_DBGPRINT(s,x,y,z) pr_info(s, x, y, z )
#else
#define KEY_DBGPRINT(s,x,y,z)
#endif 












static int button_status( void )

{
  int col;
  int row;
  unsigned int idx = 0;

  int gpio_idx = 0;  



  int key_on_num = 0;


  

  for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_col_gpios); gpio_idx++)
  {

    
    gpio_tlmm_config(GPIO_CFG(keypad_col_gpios[gpio_idx], 0, GPIO_CFG_INPUT, GPIO_CFG_PULL_UP, GPIO_CFG_2MA), GPIO_CFG_ENABLE);

  }


  
  
  
  for(col = 0; col<ARRAY_SIZE(keypad_col_gpios); col++)
  {

    
    gpio_set_value(keypad_col_gpios[col],  GPIO_LO );
    gpio_tlmm_config(GPIO_CFG(keypad_col_gpios[col], 0, GPIO_CFG_OUTPUT, GPIO_CFG_NO_PULL, GPIO_CFG_2MA), GPIO_CFG_ENABLE);

    for(row = 0; row<ARRAY_SIZE(keypad_row_gpios); row++)
    {

      key_push_flag[col][row] = KEY_RELEASE;

      if( gpio_get_value(keypad_row_gpios[row]) == GPIO_LO )
      {
        idx = KEYMAP_INDEX(row, col);
        if ( keypad_keymap[idx] != KEY_WIMAX )  
        {

          key_push_flag[col][row] = KEY_PUSH;
          count[col][row]++;
          if (count[col][row] >= BUTTONS_COUNT_THRESHOLD)
          {
            key_on_num++;
          }

        }
      }
    }
    

    gpio_tlmm_config(GPIO_CFG(keypad_col_gpios[col], 0, GPIO_CFG_INPUT, GPIO_CFG_PULL_UP, GPIO_CFG_2MA), GPIO_CFG_ENABLE);

  }


  return key_on_num;

}










static enum hrtimer_restart handle_keys(struct hrtimer *timer)
{



  bool key_on_continue = false;  

  int gpio_idx = 0;          



  int col;
  int row;
  unsigned int key_idx = 0;
  int key_on_num = 0;

  
  key_on_num = button_status();

  
  for(col = 0; col<KEYMAP_ARRAYSIZE_COLUMN; col++)
  {
    for(row = 0; row<KEYMAP_ARRAYSIZE_ROW; row++)
    {
      key_idx = KEYMAP_INDEX(row, col);
      if (key_push_flag[col][row])
      {

        
        if (key_on_num > KEY_PUSH_MULTI_MAX)
        {
          if (count[col][row] > BUTTONS_COUNT_THRESHOLD)
          {
            if(KEY_SPACE == keypad_keymap[key_idx])
            {
              if(1 == spacekey_num)
              {
                
                input_report_key(keypad_dev, keypad_keymap[key_idx], 0);
                KEY_DBGPRINT("handle_keys key release keycode:%d, key_idx:%d, count:%d\n", keypad_keymap[key_idx], key_idx, count[col][row] );
              }
              spacekey_num--;
            }
            else
            {
              
              input_report_key(keypad_dev, keypad_keymap[key_idx], 0);
              KEY_DBGPRINT("handle_keys key release keycode:%d, key_idx:%d, count:%d\n", keypad_keymap[key_idx], key_idx, count[col][row] );
            }
          }
          count[col][row] = 0;
        }
        else
        {
          if (count[col][row] == BUTTONS_COUNT_THRESHOLD)
          {

            if(keypad_keymap[key_idx] == KEY_SPACE)
            {
              
              spacekey_num++;
              if(1 == spacekey_num)
              {
                
                input_report_key(keypad_dev, keypad_keymap[key_idx], 1);

                
                if(KEYPRESS_MAX_NUM>keypad_count[key_idx])
                {
                  (keypad_count[key_idx])++;
                }
                KEY_DBGPRINT("handle_keys key press keycode:%d, key_idx:%d, count:%d\n", keypad_keymap[key_idx], key_idx, keypad_count[key_idx] );

              }
            }
            else
            {
              
              input_report_key(keypad_dev, keypad_keymap[key_idx], 1);

              
              if(KEYPRESS_MAX_NUM>keypad_count[key_idx])
              {
                (keypad_count[key_idx])++;
              }
              KEY_DBGPRINT("handle_keys key press keycode:%d, key_idx:%d, count:%d\n", keypad_keymap[key_idx], key_idx, keypad_count[key_idx] );

            }

          }

          

          key_on_continue = true;


        }

      }
      else 
      {
        if (count[col][row] >= BUTTONS_COUNT_THRESHOLD)
        {

          if(KEY_SPACE == keypad_keymap[key_idx])
          {
            if(1 == spacekey_num)
            {
              
              input_report_key(keypad_dev, keypad_keymap[key_idx], 0);
              KEY_DBGPRINT("handle_keys key release keycode:%d, key_idx:%d, count:%d\n", keypad_keymap[key_idx], key_idx, count[col][row] );
            }
            spacekey_num--;
          }
          else
          {
            
            input_report_key(keypad_dev, keypad_keymap[key_idx], 0);
            KEY_DBGPRINT("handle_keys key release keycode:%d, key_idx:%d, count:%d\n", keypad_keymap[key_idx], key_idx, count[col][row] );
          }

        }
        count[col][row] = 0;
      }
    }
  }




  if (key_on_continue) {


    
    
    hrtimer_start(&key_poll_timer, ktime_set(0, BUTTONS_POLL_INTERVAL*NSEC_PER_MSEC), HRTIMER_MODE_REL);

  }
  else {


    
    
    for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_col_gpios); gpio_idx++)
    {

      
      gpio_set_value(keypad_col_gpios[gpio_idx],  GPIO_LO );
      gpio_tlmm_config(GPIO_CFG(keypad_col_gpios[gpio_idx], 0, GPIO_CFG_OUTPUT, GPIO_CFG_NO_PULL, GPIO_CFG_2MA), GPIO_CFG_ENABLE);

    }

    
    
    for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_row_gpios); gpio_idx++)
    {
      enable_irq(gpio_to_irq(keypad_row_gpios[gpio_idx]));
    }



    
    spacekey_num = 0;

    wake_unlock(&key_wake_lock);
  }


  return HRTIMER_NORESTART;
}













irqreturn_t key_in_isr(int irq, void *ptr)
{
  int gpio_idx = 0;          


  KEY_DBGPRINT("key_in_isr start%s%s%s\n", "", "", "");

  
  for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_row_gpios); gpio_idx++)
  {
    disable_irq_nosync(gpio_to_irq(keypad_row_gpios[gpio_idx]));
  }

  
  wake_lock(&key_wake_lock);


  
  spacekey_num = 0;


  
  hrtimer_start(&key_poll_timer, ktime_set(0, 0), HRTIMER_MODE_REL);

  return IRQ_HANDLED;
}














static int keypad_gpio_event_matrix_func(struct gpio_event_input_devs *input_devs,
					  struct gpio_event_info *info,
					  void **data, int func)
{
    int err;

	int irq_err = 0;     
    int gpio_idx = 0;     


    int cnt;


    memset(&keypad_count, 0x00, sizeof(keypad_count));




    
    for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_row_gpios); gpio_idx++)
    {
      
      gpio_tlmm_config(GPIO_CFG(keypad_row_gpios[gpio_idx], 0, GPIO_CFG_INPUT, GPIO_CFG_PULL_UP, GPIO_CFG_2MA), GPIO_CFG_ENABLE);
    }

    for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_col_gpios); gpio_idx++)
    {
      
      
      
      
      
      
      gpio_set_value(keypad_col_gpios[gpio_idx],  GPIO_HI );
      gpio_tlmm_config(GPIO_CFG(keypad_col_gpios[gpio_idx], 0, GPIO_CFG_OUTPUT, GPIO_CFG_NO_PULL, GPIO_CFG_2MA), GPIO_CFG_ENABLE);
    }


    set_bit(EV_KEY, input_devs->dev[0]->evbit);
    set_bit(EV_SW, input_devs->dev[0]->evbit);
    for (cnt = 0; cnt < ARRAY_SIZE(keypad_keymap); cnt++)
    {
      set_bit(keypad_keymap[cnt], input_devs->dev[0]->keybit);
    }


    keypad_dev = input_devs->dev[0];

    hrtimer_init(&key_poll_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    key_poll_timer.function = handle_keys;


    wake_lock_init(&key_wake_lock, WAKE_LOCK_SUSPEND, "keypad");

    
    for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_row_gpios); gpio_idx++)
    {
      
	  irq_err |= request_irq(gpio_to_irq(keypad_row_gpios[gpio_idx]),   key_in_isr, IRQF_TRIGGER_LOW, "KEYSENSE_N_0", 0);

      
      enable_irq_wake(gpio_to_irq(keypad_row_gpios[gpio_idx]));
    }

	if (irq_err) {
		printk(KERN_ERR "key request_irq() failed!\n");
		return irq_err;
	}

    
    
    
    for(gpio_idx = 0; gpio_idx < ARRAY_SIZE(keypad_col_gpios); gpio_idx++)
    {
      gpio_set_value(keypad_col_gpios[gpio_idx],  GPIO_LO );
    }



    err = 0;

    return err;
}










unsigned int* get_keypress_cnt_result(int* array_len)
{
  *array_len = sizeof(keypad_count);
  return keypad_count;
}

